/*
   $Id: CommandSlice.cpp 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iostream>
#include <string>
#include <memory>
using namespace std;

#include "ptypes.h"
#include "QueryBuffer.h"
#include "QueryBufferDouble.h"

#include "ParametersSlice.h"
#include "CommandSlice.h"

#ifndef __LP64__
    #error "GLINT MAY BE COMPILED ONLY ON A 64 BITS PLATFORM"
#endif

extern RealLogger* logger;

class DisplayBAMLowLevel {
	public:
		DisplayBAMLowLevel(const bam_header_t* h): __parameters(Parameters::Instance()), __header(h), __fp(NULL),__chunk(1) {};
		void NextChunk() { __chunk++; if ( __fp != NULL ) {samclose(__fp);};__fp=NULL;};
		void WriteBuffer( const bam1_t* );
	private:
		Parameters& __parameters;
		const bam_header_t *__header;
		samfile_t* __fp;
		int __chunk;
		CompressedFileName __file_name;
};

void DisplayBAMLowLevel::WriteBuffer( const bam1_t* b )
{
	if ( __fp == NULL )
	{
		__parameters.SetChunkNb(__chunk);
		__file_name = __parameters.GetOutputFilename();
		__fp = samopen ( __file_name.GetName().c_str(), "wb",__header );
		if ( __fp == NULL )
		{
			string msg = " *** ERROR - Cannot open file " + __file_name; 
			throw ios_base::failure(msg);
		}
	}
	
	int bytes_written = samwrite(__fp,b);

	if ( bytes_written == 0 )
	{
		string msg = " *** ERROR - Cannot write to file " + __file_name; 
		throw ios_base::failure(msg);
	}
}

/*////////////// Main ///////////////*/
int CommandSlice::Main(int argc,char* argv[])
{
	try
	{
		ReverseComplement::Init();

		Parameters& parameters = Parameters::Instance();
		ParametersInitializer& prm_init = _GetInitializer();
		prm_init.Init(argc, argv, *this);

		if ( !parameters.IsSilent() ) 
			PrintHeader();
		
	    /* Initializing logger */
	    string log_filename;
	    string query_filename = parameters.GetQueryFilename().GetName();
	    if (query_filename == "-")
	    	log_filename="glint_pipe.log";
	    else
	    	log_filename=query_filename + ".log";
	    
	    auto_ptr<RealLogger> logger_auto = auto_ptr<RealLogger>(new RealLogger(log_filename,parameters.GetLogLevel(),LOGGER_OPT_AUTOFLUSH));
	    logger = logger_auto.get();
	    
	    logger->SetLogFormat(no_stamp);

		bool pair_end = parameters.IsPairEnd();

		uint64_t chunk_size = parameters.GetOutputChunkSize();
		uint64_t seq_nb   = 0;
		if ( pair_end == true )
		{
			QueryBufferBamLowLevelDouble query(query_filename);
			bam_header_t* bam_header = query.GetHeader();
			DisplayBAMLowLevel display(bam_header);
			while (query.ReadNextSeq())
			{
				const bam1_t* b = query.GetFwdLowLevelBuffer();
				display.WriteBuffer(b);
				const bam1_t* b1= query.GetBwdLowLevelBuffer();
				display.WriteBuffer(b1);
				if ( chunk_size != 0 && ++seq_nb == chunk_size )
				{
					seq_nb = 0;
					display.NextChunk();
				}
			}
		}
		else
		{
			QueryBufferBamLowLevel query(query_filename);
			bam_header_t* bam_header = query.GetHeader();
			DisplayBAMLowLevel display(bam_header);
			while (query.ReadNextSeq())
			{
				const bam1_t* b = query.GetLowLevelBuffer();
				display.WriteBuffer(b);
				if ( chunk_size != 0 && ++seq_nb == chunk_size )
				{
					seq_nb = 0;
					display.NextChunk();
				}
			}
		}
    }
   	catch ( exception& e)
	{	
		cout << e.what() << endl;
		return 1;
	};
	
	return 0;
}

/*Utilisation du programme */
void CommandSlice::Usage() const
{
	PrintHeader();
	cerr << "glint slice: Cutting a huge bam file in several \"slices\".\n\n";
	cerr << "Usage:  glint slice [--pair-end] [--chunk-size 1000] input.bam -o output.%chunk.bam \n";
	cerr << "\t\t input.bam         : The input file, bam formatted\n";
	cerr << "\t\t output.%chunk.bam : The output file, bam formatted too\n";
	cerr << "\t\t WARNING - The output file names should include the substring \"%chunk\" somewhere in their names, it will be replaced with a number !\n";
	cerr << "Optional:\n";
	cerr << "\t\t --pair-end : Input and output data are paired: A seq and its mate are alwyas in the same slice, one after onather\n";
	cerr << "\t\t --chunk-size 1000: When there are more than 1000 sequences (or 1000 pairs) to write, a new output file is opened\n";
}
