/*
   $Id
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <iostream>
#include <string>
using namespace std;

#include "utils.h"
#include "DisplaySam.h"
#include "QueryBuffer.h"
#include "Bank.h"
#include "hsp.h"
#include "SeqInfo.h"
#include "CompactTree.h"
#include "../samtools/sam.h"
#include "../samtools/bam.h"

/**
 * \brief Building an Extended Cigar representation of the alignment
 *        This object is used to build a cigar and store it directly inside a bam1_t object
 *        The AddOperation records an operation in the object.
 *        The Close() function is called to record the last operation. It is impossible to add any operation after the cigar is closed
 *        The Revert() function reverts the cigar (only on a closed cigar)
 *        The GetNbOps() method returns the number of operations, but it may only be called on an already closed cigar
 */
class Cigar
{
	public:
		Cigar(uint32_t* b): __last_operation(7),__repet(0),__count(0),__index(0),__base(b),__closed(false){};
		/**
		 * \brief Add a new operation to the cigar - Legals ops are 0,1,2,3,4,5,6, 7 means "none"
		 * \param op The operation to add
		 * \param fct the count factor (useful to add several times the same operation)
		 */
		void AddOperation( uint8_t op, uint32_t fct=1 )
		{
			if (__closed)
			{
				throw ( logic_error("ERROR - The cigar is closed") );
			}
			
			if ( op == __last_operation )
			{
				__repet += fct;
			}
			else
			{
				if (__repet > 0) // Nothing to record at first operation
					__base[__index++] = (__repet<<BAM_CIGAR_SHIFT) + (__last_operation & 0x000F);
				__repet = fct;
				__last_operation = op;
				if ( op != 7 ) // Nothing to count at last operation
					__count++;
			}
		};
		void Close()
		{
			AddOperation(7);
			__closed = true;
		};
		uint32_t GetNbOps()
		{
			if (__closed==false)
			{
				throw ( logic_error("ERROR - The cigar is NOT YET closed") );
			}
			else
			{
				return __count;
			}
		};
		void Revert()
		{
			for ( uint32_t i=0; i<__count/2; i++ )
			{
				uint32_t tmp        = __base[i];
				__base[i]           = __base[__count-1-i];
				__base[__count-1-i] = tmp;
			}
		};
	
	private:
		uint8_t   __last_operation;
		uint32_t  __repet;
		uint32_t  __count;
		uint32_t  __index;
		uint32_t* __base;
		bool      __closed;
};

/**
 * \brief The DisplaySAM constructor: not in the .h because of the __bank dependancy
 */
DisplaySAM::DisplaySAM(bool ext, bool mate): Display(),__bank(Bank::Instance()),__bam(false),__ext(ext),__mate(mate),__bank_seq(CompactTree::Instance().GetSeq()) {}
	
/**
 * \brief Display the sam header
 */

void DisplaySAM::DisplayTopHeader() const
{
	if ( !__samout.IsOpened() )
	{
		SeqInfo seq_info     = __bank.GetSeqInfo();
		bam_header_t bam_header;

		bam_header.n_targets = seq_info.size();
		bam_header.target_name = new char*[bam_header.n_targets];
		bam_header.target_len  = new uint32_t[bam_header.n_targets];
		for (int i=0; i<bam_header.n_targets; i++)
		{
			bam_header.target_name[i]    = (char*) seq_info[i].Name.c_str();
			bam_header.target_len[i]     = seq_info.GetLength(i);
			__name2tid[seq_info[i].Name] = i;
		}
		bam_header.dict   = NULL;
		bam_header.hash   = NULL;
		bam_header.rg2lib = NULL;
		bam_header.l_text = 0;
		bam_header.n_text = 0;
		bam_header.text   = NULL;
		const char* mode = (__bam) ? "wb" : "wh";
		const char* output_file;
		if (_parameters.IsOutputPipe())
		{
			output_file = "-";
		}
		else
		{
			output_file = (__mate?_parameters.GetOutputBwdFilename().GetName().c_str():_parameters.GetOutputFilename().GetName().c_str());
		}
		//__samout.SamOpen( _parameters.IsOutputPipe()?"-":_parameters.GetOutputFilename().c_str(), mode, &bam_header);
		__samout.SamOpen( output_file, mode, &bam_header );
		delete[] bam_header.target_len;
		delete[] bam_header.target_name;
	}
}

/**
 * \brief displayHsp for the SAM format
 */

void DisplaySAM::DisplayHsp(const Hsp& hsp) const
{
	bam1_t *b=bam_init1();
	b->core.tid   = __name2tid[hsp.GetSubjectName()];
	b->core.flag  = (hsp.GetStrand()==1?0:BAM_FREVERSE);	
	b->core.n_cigar = 1;
	b->core.mtid  = b->core.tid;
	b->core.mpos  = 0 ;
	b->core.bin   = bam_reg2bin(hsp.GetBQueryStart(),hsp.GetBQueryEnd());
	b->core.isize = 0;
	if ( hsp.GetScore() <= 255 )
	{
		b->core.qual  = hsp.GetMapQuality();
	}
	else
	{
		string msg = "ERROR - Your alignment is too long (score > 255) to be stored in a SAM/BAM file. Please try another output format\n";
		msg += "        query   = ";
		msg += hsp.GetQueryName();
		msg += '\n';
		msg += "        subject = ";
		msg += hsp.GetSubjectName();
		throw( runtime_error(msg));
	}
	
	b->l_aux    = 0;
	b->m_data   = 5 * hsp.GetQueryLength();

	b->core.pos   = hsp.GetSubjectStart()-1;    // 0-based coordinate
			
	b->data = (uint8_t*) malloc(b->m_data);
	if (b->data==NULL)
		throw(bad_alloc());
	else
		memset(b->data, 0, b->m_data );

	// Formatting the data of this hsp
	string readname = GetReadName( hsp.GetQueryName() );
	b->core.l_qname = readname.length() + 1; // including the trailing \0
	strcpy (bam1_qname(b),readname.c_str());
	__WriteCigar (hsp,b);
	__CopySequence (hsp,b);
	__WriteQual( hsp,b );
		
	b->data_len = (bam1_aux(b) - b->data) * sizeof ( uint8_t );

	// Writing some auxiliary data
	__WriteAux( hsp,b );
	
	// If the reads were produced in pair-end mode FragNum > 0
	b->core.flag  |= ( _query_ptr->GetFragNum()>0 )?  BAM_FPAIRED : 0;
	//If the hsp is paired writing some additional informations
	if ( hsp.IsPaired() )   
	{
		__WritePairingInfos( hsp, b );
	}
		
	// Writing this hsp to the file
	samwrite( __samout.Get(), b );
	
	// Freeing memory
	bam_destroy1(b);
}

/**
 * \brief Formats the query sequence of the hsp and stores it a bam1_t object
 * \param hsp[in]   The hsp to convert
 * \param b[in,out] The bam1_t object
 */

void DisplaySAM::__CopySequence ( const Hsp& hsp, bam1_t* b) const
{
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));

	uint8_t* d = bam1_seq(b);
	const unsigned char* s    = _query_ptr -> GetBase()+hsp.GetBQueryStart();
	uint32_t len  = hsp.GetQueryLength();
	
	// Strand + Copy the sequence
	if ( hsp.GetStrand() == 1 )
	{
		uint32_t i;
		for ( i=0; i<len; ++i )
		{
			if ( i%2 == 0 )
			{
				d[i/2] = bam_nt16_table[(int)s[i]] << 4;
			}
			else
			{
				d[i/2] += bam_nt16_table[(int)s[i]];
			}
		}
		// Finish with an \0
		if ( i%2 == 0 )
		{
			d[i/2] = 0;
		}
		b->core.l_qseq  = hsp.GetQueryLength();
	}
	// Strand - Copy the reverse sequence
	else
	{
		ReverseComplement& rev = ReverseComplement::Instance();
		if ( len != 0 )
		{
			uint32_t i=len;
			uint32_t j=0;
			do
			{
				--i;
				int c = rev[s[i]];
				if ( j%2 == 0 )
				{
					d[j/2] = bam_nt16_table[c] << 4;
				}
				else
				{
					d[j/2] += bam_nt16_table[c];
				}
				++j;
			} while ( i!=0 );
			// Finish with an \0
			if ( j%2 == 0 )
			{
				d[j/2] = 0;
			}
			b->core.l_qseq  = hsp.GetQueryLength();
		}
	}
}

/**
 * \brief Writes the quality area, initializing it to ascii 33 (base quality)
 *        If the hsp is on the reverse strand, we reverse the quality (as the query is reverse & complemented)
 * 
 * \param len[int] The sequence length
 * \param b       The bam1_t object
 * 
 */
void DisplaySAM::__WriteQual ( const Hsp& hsp, bam1_t* b) const
{
	if ( _query_ptr == NULL )
	{
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
	}
	else
	{
		if ( _query_ptr -> GetQual()==NULL )
		{
			uint8_t* d = bam1_qual(b);
			for ( uint32_t i=0; i < hsp.GetQueryLength(); ++i )
			{
				*d++ = 0;
			}
		}
		else
		{
			uint8_t* dst = bam1_qual(b);
			size_t len   = hsp.GetQueryLength();
			memcpy ( (void*) dst, (void*) _query_ptr -> GetQual(),len );
			if ( hsp.GetStrand() == -1 )
			{
				Reverse((unsigned char*) dst, (uint32_t) len);
			}
		}
	}
}


/**
 * \brief A little wrapper for bam_aux_del: can be called specifying a tag, works even if the tag does not exist
 */
void BamAuxDel ( bam1_t *b, const char tag[2] )
{
	uint8_t* s = bam_aux_get(b,tag);
	if ( s!=NULL)
	{
		bam_aux_del(b,s);
	}
}

/**
 * \brief Write auxiliary data: 
 *        - 1st write auxliary data from the input, if any
 *        - 2nd delete tags related to a previous alignment and write NM tag
 */
void DisplaySAM::__WriteAux ( const Hsp& hsp, bam1_t* b) const
{
	if ( _query_ptr == NULL )
	{
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
	}
	else
	{
		if ( _query_ptr -> GetAuxData() != NULL )
		{
			b->data_len += _query_ptr -> GetAuxDataLength();
			b->l_aux    =  _query_ptr -> GetAuxDataLength();
			if (b->m_data < b->data_len) {
				b->m_data = b->data_len;
				kroundup32(b->m_data);
				b->data = (uint8_t*)realloc(b->data, b->m_data);
			}
			memcpy ( bam1_aux(b), _query_ptr -> GetAuxData(), _query_ptr -> GetAuxDataLength() );
			BamAuxDel(b, "MD");
			BamAuxDel(b, "H0");
			BamAuxDel(b, "H1");
			BamAuxDel(b, "H2");
			BamAuxDel(b, "NM");
			BamAuxDel(b, "AS");
			BamAuxDel(b, "XA");
			BamAuxDel(b, "XO");
			BamAuxDel(b, "X0");
			BamAuxDel(b, "X1");
		}
	}

	int edit_dist = hsp.GetEditDist();
	int32_t score   = hsp.GetScore();
	int gap_openings = (int) hsp.GetGapOpenings().GetSize();
	bam_aux_append(b, "NM", 'i', sizeof(int),   (uint8_t*)&edit_dist);
	bam_aux_append(b, "AS", 'i', sizeof(int32_t), (uint8_t*)&score);
	bam_aux_append(b, "XO", 'i', sizeof(int), (uint8_t*)&gap_openings);
			
	if ( __ext )
	{
		__WriteMD(hsp,b);
	}
}

/**
 * \brief Write additional pairing information: 
 *       
 */
//TODO Modify mapping quality in pair-end mode !! (see static int pairing( ) in bwape.c)
void DisplaySAM::__WritePairingInfos ( const Hsp& hsp, bam1_t* b) const
{
	HspPtr mate_ptr =  hsp.GetMate();
	
	b->core.mtid  = __name2tid[mate_ptr->GetSubjectName()];
	b->core.mpos  = mate_ptr->GetSubjectStart()-1;    // 0-based coordinate	
	b->core.isize = ( hsp.GetStrand() == 1 ) ?  mate_ptr->GetSubjectEnd() - hsp.GetSubjectStart() + 1 : hsp.GetSubjectEnd() - mate_ptr->GetSubjectStart()+ 1 ;
	b->core.flag|= BAM_FPROPER_PAIR;
	b->core.flag|= ( _query_ptr->GetFragNum() == FWD ) ? BAM_FREAD1 : BAM_FREAD2;
		
	if ( mate_ptr->GetStrand() == -1 )
		b->core.flag|= BAM_FMREVERSE;	

	MapInfo mapinfo = hsp.GetMapInfo();	
	for ( MapInfoIterator i =  mapinfo.begin(); i!=  mapinfo.end(); ++i )
	{		
		string s = i->first;
		//cerr<<"here "<<s<<endl;
		bam_aux_append(b, s.c_str(), 'i', sizeof(int), (uint8_t*) &i->second );	
	}
}
		
/**
 * \brief Computes the extended cigar value from the hsp, and stores this value in a bam1_t object
 * \param hsp[in]   The hsp to convert
 * \param b[in,out] The bam1_t object
 */ 
void DisplaySAM::__WriteCigar( const Hsp& hsp, bam1_t* b ) const
{
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
	
	Cigar cigar(bam1_cigar(b));
	
	const unsigned char *nQuery =_query_ptr -> GetBase()+hsp.GetBQueryStart();
	const unsigned char *nBase  = (hsp.GetStrand()>0) ?  __bank_seq+hsp.GetBSubjectStart() : __bank_seq+hsp.GetBSubjectEnd();

    GapPile_const_iterator qgap = hsp.GetQueryGaps().begin();
    GapPile_const_iterator sgap = hsp.GetSubjectGaps().begin();

	uint32_t leading_hard_clips = hsp.GetQueryStart() - 1;
	uint32_t trailing_hard_clips= _query_ptr -> GetSeqLength() - hsp.GetQueryEnd();
	
	if ( leading_hard_clips != 0 )
	{
		cigar.AddOperation( BAM_CHARD_CLIP, leading_hard_clips );
	}
	for ( uint32_t l=0; l < hsp.GetSize(); ++l )
	{
		if (qgap!=hsp.GetQueryGaps().end() && *qgap==l) {
			cigar.AddOperation(BAM_CDEL);
			++qgap;
			nBase+=hsp.GetStrand();
		}
		else if (sgap!=hsp.GetSubjectGaps().end() && *sgap==l) {
			cigar.AddOperation(BAM_CINS);
			++sgap;
			++nQuery;
		} else {
			cigar.AddOperation(BAM_CMATCH);
			nBase+=hsp.GetStrand();
			++nQuery;
		}
	}
	if ( trailing_hard_clips != 0 )
	{
		cigar.AddOperation( BAM_CHARD_CLIP, trailing_hard_clips );
	}
	cigar.Close();
	if ( hsp.GetStrand() != 1 )
	{
		cigar.Revert();
	}
	b->core.n_cigar = cigar.GetNbOps();
}

/**
 * \brief Compute the extended field MD, which gives the possibility to reconstruct the subject
 *        For a reverse strand, reverse the gaps and the mismatches, then 
 * \param hsp[in]   The hsp
 * \param b[in,out] The bam1_t object
 */ 
void DisplaySAM::__WriteMD( const Hsp& hsp, bam1_t* b ) const
{
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));

	ostringstream result;
	enum state_t { GAP, MATCH, MISMATCH };
	state_t state = MATCH;	
	int matches     = 0;
	
	const unsigned char *nBase  = __bank_seq+hsp.GetBSubjectStart();

	GapPile qgaps = (hsp.GetStrand()==1?hsp.GetQueryGaps()  :hsp.GetQueryGaps().Reverse(hsp.GetSize()));
    GapPile msmas = (hsp.GetStrand()==1?hsp.GetMismatch()   :hsp.GetMismatch().Reverse(hsp.GetSize()));
    GapPile sgaps = (hsp.GetStrand()==1?hsp.GetSubjectGaps():hsp.GetSubjectGaps().Reverse(hsp.GetSize()));

    GapPile_const_iterator qgap = qgaps.begin();
	GapPile_const_iterator msma = msmas.begin();
	GapPile_const_iterator sgap = sgaps.begin();

	/* FOR DEBUG ONLY
	string b1= (char*) nBase;
	
	string q1= (char*)(_query_ptr -> GetSeq()+hsp.GetBQueryStart());
	
	ReverseComplement& rev = ReverseComplement::Instance();
	unsigned char* q2_ptr = (unsigned char*) (_query_ptr -> GetSeq()+hsp.GetBQueryStart());
	rev(q2_ptr,hsp.GetSize());
	string q2 = (char*)(_query_ptr -> GetSeq()+hsp.GetBQueryStart());
	
	cerr << q1.substr(0,hsp.GetSize()-2) << "\n";
	cerr << q2.substr(2,hsp.GetSize()-2) << "\n";
	cerr << b1.substr(0,hsp.GetSize()) << "\n";
	cerr << "qgap ";
	for (GapPile_const_iterator i = qr_gaps.begin(); i != qr_gaps.end(); ++i ) cerr << *i << " ";
	cerr << "\nmismatches\n";
	for (GapPile_const_iterator i = msmas.begin(); i != msmas.end(); ++i ) cerr << *i << " ";
	cerr << " REVERSED => ";
	for (GapPile_const_iterator i = r_msmas.begin(); i != r_msmas.end(); ++i ) cerr << *i << " ";
	cerr << '\n';
	*/

	for ( uint32_t l=0; l < hsp.GetSize(); ++l )
	{
		// a gap in the query (thus an insertion in the subject)
		if (qgap!=qgaps.end() && *qgap==l) 
		{
			// Print the previous match count then reset it
			if ( state == MATCH )
			{
				result << matches;
				matches = 0;
			}
			// Mark the gap opening
			if ( state != GAP )
			{
				state = GAP;
				result << '^';
			}
			result << *nBase;
			++qgap;
			nBase++;
		}
		
		// a gap in the subject (thus an insertion in the query)
		else if (sgap!=sgaps.end() && *sgap==l) 
		{
			// nothing to write, we increment the sgap iterator and skip the position
			++sgap;
		}
		
		// a mismatch
		else if (msma!=msmas.end() && *msma==l) 
		{
			// Print the previous match count then reset it
			if ( state == MATCH )
			{
				result << matches;
				matches = 0;
				state = MISMATCH;
			}
			result << *nBase;
			nBase++;
			++msma;
		} 
		
		// a match - just counting the matches
		else 
		{
			state = MATCH;
			matches++;
			nBase++;
		}
	}
	result << matches;

	size_t size = sizeof(char) * (result.str().size()+1); // do not forget the \0
	bam_aux_append(b, "MD", 'Z', sizeof(char) * size, (uint8_t*)result.str().c_str());
}
